import re
import hashlib
import requests
import logging
from typing import Union
from pathlib import Path
import typer
import petl as etl
from rich import print
from requests.adapters import HTTPAdapter, Retry
import importlib.metadata


logging.basicConfig(level=logging.ERROR, format='(%(levelname)s): %(message)s')
logger = logging.getLogger(__name__)


def get_version() -> str:
    return importlib.metadata.version('gwas-sumstats-tools')


def exit_if_no_data(table: Union[etl.Table, None]) -> None:
    if table is None:
        print("No data in table. Exiting.")
        raise typer.Exit()


def parse_accession_id(filename: Path) -> Union[str, None]:
    """Get accession ID from file path

    Arguments:
        filename -- Input file path 

    Returns:
        GCST or None
    """
    gcst_search = re.search(r"GCST[0-9]+", filename.stem)
    return gcst_search.group() if gcst_search else None


def parse_genome_assembly(filename: Path) -> Union[str, None]:
    """Get genome assembly from file path

    Arguments:
        filename -- Input file path 

    Returns:
        Genome assembly or None
    """
    gcst_search = re.search(r"(build|grch)([0-9]+)", filename.stem, re.IGNORECASE)
    return gcst_search.group(2) if gcst_search else None


def download_with_requests(url,
                           params: dict = None,
                           headers: dict = None) -> Union[bytes, None]:
    """Download content from URL

    Arguments:
        url -- request url
        params -- requests parameters
        headers -- request headers

    Return:
        Content from URL if status code is 200 or None
    """
    s = requests.Session()
    retries = Retry(total=5, backoff_factor=2, status_forcelist=[429, 500, 502, 503, 504])
    s.mount(url, HTTPAdapter(max_retries=retries))
    try:
        r = s.get(url, params=params, headers=headers)
        status_code = r.status_code
        if status_code == 200:
            return r.content
        else:
            logger.error(f"{url} returned {status_code} status code")
            return None
    except requests.exceptions.RequestException as e:
        logger.error(e)
        return None


def append_to_path(path: Path, to_add: str) -> Path:
    return path.parent / (path.name + to_add)


def set_metadata_outfile_name(data_outfile: str, metadata_outfile: str) -> str:
    if metadata_outfile is None:
        metadata_outfile = data_outfile + "-meta.yaml"
    return metadata_outfile


def get_md5sum(file: Path) -> str:
    hash_md5 = hashlib.md5()
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def header_dict_from_args(args: list) -> dict:
    """Generate a dict from cli args split on ":"

    Arguments:
        args -- cli args list

    Returns:
        Dict of key, values
    """
    header_dict = {}
    for arg in args:
        if ":" not in arg:
            # skip because it's not a metadata mapping
            pass
        else:
            key, value = arg.replace("--", "").split(":")
            header_dict[key] = value
    return header_dict


def metadata_dict_from_args(args: list) -> dict:
    """Generate a dict from cli args split on "="

    Arguments:
        args -- cli args list

    Returns:
        Dict of key, values
    """
    meta_dict = {}
    for arg in args:
        if "=" not in arg:
            # skip because it's not a metadata mapping
            pass
        else:
            key, value = arg.replace("--", "").split("=")
            meta_dict[key] = value
    return meta_dict


def replace_dictionary_keys(data_dict: dict, replace_dict: dict) -> dict:
    """Replace data_dict keys with values from replace_dict

    Arguments:
        data_dict -- dict to replace keys in
        replace_dict -- dict mapping 'find' to 'replace' keys

    Returns:
        dict with keys replaced
    """
    return {replace_dict.get(k, k): v for k, v in data_dict.items()}


def update_dict_if_not_set(data_dict: dict, field: str, value: any) -> dict:
    if data_dict.get(field) is None:
        data_dict[field] = value
    return data_dict


def split_fields_on_delimiter(data_dict: dict,
                              fields: tuple,
                              delimiter: str = "|") -> dict:
    """Split specified fields in dict on delimiter

    Arguments:
        data_dict -- dict to split fields in
        fields -- fields to split

    Keyword Arguments:
        delimiter -- delimiter (default: {"|"})

    Returns:
        data_dict with fields split
    """
    return dict((k, v.split(delimiter))
                if k in fields
                else (k, v)
                for k, v
                in data_dict.items())



import os


def _env_variable_else(env_var_name: str, default: str) -> str:
    """Get an environment variable

    Arguments:
        env_var_name -- variable name
        default -- default value

    Returns:
        value of environment variable
    """
    value = os.environ.get(env_var_name)
    return value if value else default


GWAS_CAT_API_STUDIES_URL = _env_variable_else(
    "GWAS_CAT_API_STUDIES_URL", "https://www.ebi.ac.uk/gwas/rest/api/studies/"
)

GWAS_CAT_API_INGEST_STUDIES_URL = _env_variable_else(
    "GWAS_CAT_API_INGEST_STUDIES_URL",
    "https://www.ebi.ac.uk/gwas/ingest/api/v2/studies/",
)

GWAS_CAT_STUDY_MAPPINGS = {
    "genotyping_technology": "genotyping_technology",
    "traitDescription": "trait_description",
    "effect_allele_frequency_lower_limit": "minor_allele_freq_lower_limit",
    "minor_allele_frequency_lower_limit": "minor_allele_freq_lower_limit",
    "summary_statistics_assembly": "genome_assembly",
    "analysisSoftware": "analysis_software",
    "imputationPanel": "imputation_panel",
    "imputationSoftware": "imputation_software",
    "adjustedCovariates": "adjusted_covariates",
    "ontologyMapping": "ontology_mapping",
    "readme_file": "author_notes",
    "readme_text": "author_notes",
    "coordinateSystem": "coordinate_system",
    "sex": "sex",
}

GWAS_CAT_SAMPLE_MAPPINGS = {
    "size": "sample_size",
    "ancestry": "sample_ancestry",
    "ancestryMethod": "ancestry_method",
    "caseControlStudy": "case_control_study",
    "caseCount": "case_count",
    "controlCount": "control_count",
}

GENOME_ASSEMBLY_MAPPINGS = {"36": "GRCh36", "37": "GRCh37", "38": "GRCh38"}

STUDY_FIELD_TO_SPLIT = (
    "genotyping_technology",
    "trait_description",
    "ontology_mapping",
    "adjusted_covariates",
)

SAMPLE_FIELD_TO_SPLIT = ("ancestry_method", "sample_ancestry")



from pathlib import Path
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn

from gwas_sumstats_tools.interfaces.data_table import SumStatsTable
from gwas_sumstats_tools.interfaces.metadata import (
    MetadataClient,
    metadata_dict_from_gwas_cat,
    get_file_metadata,
)
from gwas_sumstats_tools.utils import (
    parse_accession_id,
    append_to_path,
    exit_if_no_data,
)


class Formatter:
    def __init__(
        self,
        data_infile: Path,
        data_outfile: Path = None,
        metadata_infile: Path = None,
        metadata_outfile: Path = None,
        format_data: bool = False,
    ) -> None:
        self.format_data = format_data
        self.data_infile = Path(data_infile)
        self.data_outfile = Path(
            self._set_data_outfile_name() if not data_outfile else data_outfile
        )
        self.metadata_outfile = Path(
            self._set_metadata_outfile_name()
            if not metadata_outfile
            else metadata_outfile
        )
        self.data = (
            SumStatsTable(sumstats_file=self.data_infile) if self.data_infile else None
        )
        self.meta = MetadataClient(
            in_file=metadata_infile, out_file=self.metadata_outfile
        )

    def _set_data_outfile_name(self) -> str:
        """Set the data outfile name.
        This method only runs if the data outfile is not
        set. Then if the data is not to be reformatted, the
        outfile is the same as the infile. If the data is
        to be formatted, the outfile name is configured based
        on infile name accession ID parsing.

        Returns:
            data outfile name string
        """
        if self.format_data:
            accession_id = parse_accession_id(filename=self.data_infile)
            if accession_id:
                self.data_outfile = accession_id + ".tsv.gz"
            else:
                self.data_outfile = append_to_path(
                    self.data_infile, "-FORMATTED.tsv.gz"
                )
        else:
            self.data_outfile = self.data_infile
        return self.data_outfile

    def _set_metadata_outfile_name(self) -> str:
        return append_to_path(self.data_outfile, "-meta.yaml")

    def set_metadata(
        self, from_gwas_cat: bool = False, custom_metadata: dict = None
    ) -> MetadataClient:
        """Set metadata.
        The hierarchy of where metadata is set is as follows:
        1. custom_metadata map (overwrites anything below)
        2. metadata from the GWAS Catalog
        3. metadata from the original input file
        4. metadata inferred from the datafile

        Keyword Arguments:
            from_gwas_cat -- update with data from GWAS catalog (default: {False})
            custom_metadata -- Update with this map (default: {None})

        Returns:
            metadata object
        """
        self.meta.from_file()
        meta_dict = get_file_metadata(
            in_file=self.data_infile,
            out_file=self.data_outfile,
            meta_dict=self.meta.as_dict(),
        )
        if from_gwas_cat:
            accession_id = parse_accession_id(filename=self.data_infile)
            meta_dict.update(metadata_dict_from_gwas_cat(accession_id=accession_id))
        if custom_metadata:
            meta_dict.update(custom_metadata)
        self.meta.update_metadata(meta_dict)
        return self.meta


def format(
    filename: Path,
    data_outfile: Path = None,
    minimal_to_standard: bool = False,
    generate_metadata: bool = False,
    metadata_outfile: Path = None,
    metadata_infile: Path = None,
    metadata_from_gwas_cat: bool = False,
    header_map: dict = None,
    metadata_dict: dict = None,
) -> None:
    formatter = Formatter(
        data_infile=filename,
        data_outfile=data_outfile,
        metadata_infile=metadata_infile,
        metadata_outfile=metadata_outfile,
        format_data=minimal_to_standard,
    )
    if minimal_to_standard:
        exit_if_no_data(table=formatter.data.sumstats)
        print("[bold]\n-------- SUMSTATS DATA --------\n[/bold]")
        print(formatter.data.sumstats)
        if header_map:
            formatter.data.reformat_header(header_map=header_map)
        else:
            formatter.data.reformat_header()
        formatter.data.normalise_missing_values()
        print("[bold]\n-------- REFORMATTED DATA --------\n[/bold]")
        print(formatter.data.sumstats)
        print(
            f"[green]Formatting and writing sumstats data --> {str(formatter.data_outfile)}[/green]"
        )
        with Progress(
            SpinnerColumn(finished_text="Complete!"),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task(description="Processing...", total=None)
            formatter.data.to_file(outfile=formatter.data_outfile)
    # Get metadata
    if generate_metadata:
        print("[bold]\n---------- METADATA ----------\n[/bold]")
        metadata = formatter.set_metadata(
            from_gwas_cat=metadata_from_gwas_cat, custom_metadata=metadata_dict
        )
        print(metadata)
        print(f"[green]Writing metadata --> {str(metadata._out_file)}[/green]")
        metadata.to_file()
    if not any([minimal_to_standard, generate_metadata]):
        print("Nothing to do.")



import yaml
import json
from typing import Union
from datetime import date
from pathlib import Path
from pydantic import ValidationError
from gwas_sumstats_tools.config import (GWAS_CAT_API_STUDIES_URL,
                                        GWAS_CAT_API_INGEST_STUDIES_URL,
                                        GWAS_CAT_STUDY_MAPPINGS,
                                        STUDY_FIELD_TO_SPLIT,
                                        SAMPLE_FIELD_TO_SPLIT,
                                        GWAS_CAT_SAMPLE_MAPPINGS,
                                        GENOME_ASSEMBLY_MAPPINGS)
from gwas_sumstats_tools.utils import (download_with_requests,
                                       parse_accession_id,
                                       parse_genome_assembly,
                                       get_md5sum,
                                       replace_dictionary_keys,
                                       split_fields_on_delimiter,
                                       update_dict_if_not_set)
from gwas_sumstats_tools.schema.metadata import SumStatsMetadata


class MetadataClient:
    def __init__(self, meta_dict: dict = {},
                 in_file: Path = None,
                 out_file: Path = None) -> None:
        """SumStats Metadata client

        Keyword Arguments:
            meta_dict -- Dict of metadata (default: {None})
            in_file -- Input metadata YAML file (default: {None})
            out_file -- Output metadata YAML file (default: {None})
        """
        self.metadata = SumStatsMetadata.construct()
        self._meta_dict = meta_dict
        self._in_file = in_file
        self._out_file = out_file

    def from_file(self) -> None:
        """Create metadata from YAML file
        """
        if self._in_file is None:
            return None
        with open(self._in_file, "r") as fh:
            self._meta_dict = yaml.safe_load(fh)
            self.update_metadata(self._meta_dict)

    def to_file(self) -> None:
        """Write metadata to YAML file
        """
        with open(self._out_file, "w") as fh:
            yaml.dump(self.metadata.dict(exclude_none=True),
                      fh,
                      encoding='utf-8')

    def update_metadata(self, data_dict: dict) -> None:
        """Create a copy of the model and update (no validation occurs)

        Arguments:
            data_dict -- Dict of data to populate model
        """
        self._meta_dict.update(data_dict)
        try:
            self.metadata = self.metadata.parse_obj(self._meta_dict)
        except ValidationError as e:
            print(f"Metadata not updated due to the following error:\n {e}\n")

    def __repr__(self) -> str:
        """Representation of metadata.
        """
        return self.as_yaml()

    def as_dict(self) -> dict:
        """Dict repr of metadata

        Returns:
            Dict of metadata
        """
        return self.metadata.dict()

    def as_yaml(self, **kwargs) -> str:
        return yaml.dump(self.metadata.dict(**kwargs))


def metadata_dict_from_gwas_cat(accession_id: str) -> dict:
    """Extract metadat from the GWAS Catalog API

    Arguments:
        accession_id -- GWAS Catalog accession ID

    Returns:
        Metadata dict
    """
    study_url = GWAS_CAT_API_INGEST_STUDIES_URL + accession_id
    study_response = download_with_requests(url=study_url)
    meta_dict = _parse_gwas_api_study_response(study_response,
                                         replace_dict=GWAS_CAT_STUDY_MAPPINGS,
                                         fields_to_split=STUDY_FIELD_TO_SPLIT)
    sample_url = study_url + "/samples"
    sample_response = download_with_requests(url=sample_url, params={"size": 100})
    samples_list = _parse_gwas_api_samples_response(sample_response,
                                                    replace_dict=GWAS_CAT_SAMPLE_MAPPINGS,
                                                    fields_to_split=SAMPLE_FIELD_TO_SPLIT)
    meta_dict['samples'] = samples_list
    return meta_dict


def _parse_gwas_api_study_response(response: bytes,
                                   replace_dict: dict = None,
                                   fields_to_split: tuple = None
                                   ) -> dict:
    """Parse study repsonse from GWAS cat api

    Arguments:
        response -- response bytes

    Keyword Arguments:
        replace_dict -- Header mappings (default: {None})
        fields_to_split -- fields to split (default: {None})

    Returns:
        Dict of metadata
    """
    result_dict = {}
    if response:
        result_dict = json.loads(response.decode())
        if replace_dict:
            result_dict = replace_dictionary_keys(data_dict=result_dict,
                                                  replace_dict=replace_dict)
        if fields_to_split:
            result_dict = split_fields_on_delimiter(data_dict=result_dict,
                                                    fields=fields_to_split)
    return result_dict


def _parse_gwas_api_samples_response(response: bytes,
                                     replace_dict: dict = None,
                                     fields_to_split: tuple = None
                                     ) -> list:
    """Parse the samples response from GWAS cat api

    Arguments:
        response -- responce bytes

    Keyword Arguments:
        replace_dict -- Header mappings (default: {None})
        fields_to_split -- fields to split (default: {None})

    Returns:
        List of samples dicts
    """
    formatted_list = []
    if response:
        result_dict = json.loads(response.decode())
        sample_list = result_dict["_embedded"].get('samples')
        if sample_list:
            for element in sample_list:
                if replace_dict:
                    element = replace_dictionary_keys(data_dict=element,
                                                      replace_dict=replace_dict)
                if fields_to_split:
                    element = split_fields_on_delimiter(data_dict=element,
                                                        fields=fields_to_split)
                formatted_list.append(element)
    return formatted_list


def get_file_metadata(in_file: Path, out_file: str, meta_dict: dict = {}) -> dict:
    """Get file related metadata

    Arguments:
        in_file -- sumstats in file
        outfile -- sumstats out file

    Returns:
        Metadata dict
    """
    inferred_meta_dict = {}
    inferred_meta_dict['gwas_id'] = parse_accession_id(filename=in_file)
    inferred_meta_dict['data_file_name'] = Path(out_file).name
    inferred_meta_dict['file_type'] = 'GWAS-SFF v1.0'
    inferred_meta_dict['genome_assembly'] = GENOME_ASSEMBLY_MAPPINGS.get(parse_genome_assembly(filename=in_file), 'unknown')
    inferred_meta_dict['data_file_md5sum'] = get_md5sum(out_file) if Path(out_file).exists() else None
    inferred_meta_dict['date_last_modified'] = date.today()
    inferred_meta_dict['gwas_catalog_api'] = GWAS_CAT_API_STUDIES_URL + parse_accession_id(filename=in_file)
    for field, value in inferred_meta_dict.items():
        update_dict_if_not_set(meta_dict, field, value)
    return meta_dict


def init_metadata_from_file(filename: Path, metadata_infile: Path = None) -> Union[SumStatsMetadata, None]:
    m_in = metadata_infile if metadata_infile else filename.with_suffix(filename.suffix + "-meta.yaml")
    if m_in.exists():
        ssm = MetadataClient(in_file=m_in)
        ssm.from_file()
        return ssm
    else:
        return None


"""
Pandera Schema https://pandera.readthedocs.io for defining 
the summary statistics data tables. The schema is dynamically
generated because of columns that can be defined in multiple
ways e.g. the effect size column (index 4) can be a beta, an
odds ratio or a hazard ratio - and these all have differing 
validation constraints.
"""

from collections import OrderedDict
from pandera import Column, DataFrameSchema, Check
from pandera.dtypes import Float128

import platform

class SumStatsSchema:
    """Pandera DataFrameSchema interface for
    summary statistics data.
    
    Note on the pvalue validator:
    
    Choice of standard allowing zero or 
    -log10 allowing zero. The zero constraint
    is applied to the mantissa.
    The float type is Float128 but even
    this is not precise enough for some
    datasets, whose very small values evaluate
    to 0, if we don't split mantissa and exp.
    """
    EFFECT_FIELD_DEFINITIONS = {
        "beta": Column(float),
        "odds_ratio": Column(float, [
            Check.ge(0,
                     error="Must be a value greater than or equal to 0")
            ]),
        "hazard_ratio": Column(float, [
            Check.ge(0,
                     error="Must be a value greater than or equal to 0")
            ])
        }
    PVALUE_FIELD_DEFINITIONS = {
        'p_value': Column(float if platform.system() == "Windows" else Float128, [
            Check.in_range(0, 1,
                           include_min=True,
                           error="Must be a value between 0 and 1, inclusive of 0")
            ]),
        'neg_log_10_p_value': Column(float, [
            Check.ge(0,
                     error="Must be greater than or equal to 0")
            ])
    }
    MANTISSA_VALIDATORS = {
        'default': Column(float, [
            Check.gt(0,
                     error="Must be greater than 0")
            ]),
        'pval_zero': Column(float, [
            Check.ge(0,
                     error="Must be greater than or equal to 0")
            ])
    }
    FILE_EXTENSIONS = {
        ".tsv",
        ".tsv.gz"
    }

    def __init__(self,
                 effect_field: str = 'beta',
                 pval_field: str = 'p_value',
                 pval_zero: bool = False) -> None:
        self.effect_field = effect_field
        self.pval_field = pval_field
        self.pval_zero = pval_zero

    def schema(self) -> DataFrameSchema:
        all_fields = OrderedDict(self.mandatory_fields(),
                                 **self.optional_fields())
        schema = DataFrameSchema(
            all_fields,
            coerce=True
        )
        return schema

    def mandatory_fields(self) -> dict:
        return OrderedDict({
                "chromosome": Column(int, [
                    Check.in_range(1, 25,
                                   error="Must be a value between 1 and 25"),
                    ]),
                "base_pair_location": Column(int, [
                    Check.ge(0,
                             error="Must be positive integer")
                    ]),

                "effect_allele": Column(str, [
                    Check.str_matches(r'^LONG_STRING$|^[ACTGactg]+$',
                                      error="Must be nucleotide sequence")
                    ]),
                "other_allele": Column(str, [
                    Check.str_matches(r'^LONG_STRING$|^[ACTGactg]+$',
                                      error="Must be nucleotide sequence")
                    ]),
                self.effect_field: self.EFFECT_FIELD_DEFINITIONS.get(self.effect_field),
                "standard_error": Column(float),
                "effect_allele_frequency": Column(float, [
                    Check.in_range(0, 1,
                                   error="Must be a value between 0 and 1, inclusive")
                    ]),
                self.pval_field: self.PVALUE_FIELD_DEFINITIONS.get(self.pval_field),
                "_p_value_mantissa": self._get_mantissa_validator(),
                "_p_value_exponent": Column("Int64", nullable=True)           
            })

    def field_order(self) -> tuple:
        return tuple(k for k in self.mandatory_fields() if not k.startswith("_"))

    def optional_fields(self) -> dict:
        return {
                "variant_id": Column(str, [
                    Check.str_matches(r'^[A-Za-z0-9_]+$',
                                      error="Must match pattern")
                    ], nullable=True, required=False),
                "rsid": Column(str, [
                    Check.str_matches(r'^rs[0-9]+$',
                                      error="Must match rsID pattern")
                    ], nullable=True, required=False),
                "ref_allele": Column(str, [
                    Check.isin(('OA', 'EA'))
                    ], nullable=True, required=False),
                "ci_upper": Column(float, nullable=True, required=False),
                "ci_lower": Column(float, nullable=True, required=False),
                "info": Column(float, [
                    Check.in_range(0, 1,
                                   error="Must be a value between 0 and 1, inclusive")
                    ], nullable=True, required=False),
                "n": Column("Int64", [
                    Check.ge(0,
                             error="Must be greater than or equal to 0")
                    ], nullable=True, required=False)
            }

    def _get_mantissa_validator(self) -> Column:
        """Mantissa validator

        Returns:
            _description_
        """
        if self.pval_zero:
            # Constaint for zero value pvalues here
            return self.MANTISSA_VALIDATORS.get('pval_zero')
        else:
            return self.MANTISSA_VALIDATORS.get('default')



"""
A pydantic (https://docs.pydantic.dev/) schema that can be 
used by python projects for defining the metadata model.
"""

from pydantic import (BaseModel,
                      constr)
from datetime import date
from typing import List, Optional
from enum import Enum

"""
Enums
"""


class SexEnum(str, Enum):
    male = 'M'
    female = 'F'
    combined = 'combined'


class CoordinateSystemEnum(str, Enum):
    zero = '0-based'
    one = '1-based'


"""
Models
"""


class SampleMetadata(BaseModel):
    ancestry_method: Optional[List[str]] = None
    case_control_study: Optional[bool] = None
    case_count: Optional[int] = None
    control_count: Optional[int] = None
    sample_size: int = None
    sample_ancestry: List[str] = None


class SumStatsMetadata(BaseModel):
    genotyping_technology: List[str] = None
    gwas_id: Optional[constr(regex=r'^GCST\d+$')] = None
    trait_description: List[str] = None
    minor_allele_freq_lower_limit: Optional[float] = None
    data_file_name: str = None
    file_type: str = None
    data_file_md5sum: str = None
    is_harmonised: Optional[bool] = False
    is_sorted: Optional[bool] = False
    date_last_modified: date = None
    genome_assembly: str = None
    analysis_software: Optional[str] = None
    imputation_panel: Optional[str] = None
    imputation_software: Optional[str] = None
    harmonisation_reference: Optional[str] = None
    adjusted_covariates: Optional[List[str]] = None
    ontology_mapping: Optional[List[str]] = None
    author_notes: Optional[str] = None
    gwas_catalog_api: Optional[str] = None
    sex: Optional[SexEnum] = None
    coordinate_system: CoordinateSystemEnum = None
    samples: List[SampleMetadata] = None

    class Config:
        title = 'GWAS Summary Statistics metadata schema'
        use_enum_values = True